"""
PVs
# Last update Jan 2021 by RLO
# Now with Eiger goodness
"""

requested_transmission = "BL24I-OP-ATTN-01:T2A:SETVAL1"


# Eiger
eiger_filepath = "BL24I-EA-EIGER-01:CAM:FilePath"
eiger_filename = "BL24I-EA-EIGER-01:CAM:FileName"
eiger_od_filepath = "BL24I-EA-EIGER-01:OD:FilePath"
eiger_od_filename = "BL24I-EA-EIGER-01:OD:FileName"
eiger_seq_id = "BL24I-EA-EIGER-01:CAM:SequenceId"
eiger_numimages = "BL24I-EA-EIGER-01:CAM:NumImages"
eiger_od_num_capture = "BL24I-EA-EIGER-01:OD:NumCapture"
eiger_numexpimage = "BL24I-EA-EIGER-01:CAM:NumExposures"
eiger_acquiretime = "BL24I-EA-EIGER-01:CAM:AcquireTime"
eiger_acquireperiod = "BL24I-EA-EIGER-01:CAM:AcquirePeriod"
eiger_imagemode = "BL24I-EA-EIGER-01:CAM:ImageMode"
eiger_triggermode = "BL24I-EA-EIGER-01:CAM:TriggerMode"
eiger_numtriggers = "BL24I-EA-EIGER-01:CAM:NumTriggers"
eiger_manualtrigger = "BL24I-EA-EIGER-01:CAM:ManualTrigger"
eiger_trigger = "BL24I-EA-EIGER-01:CAM:Trigger"
eiger_filewriter = "BL24I-EA-EIGER-01:CAM:FWEnable"
eiger_stream = "BL24I-EA-EIGER-01:CAM:StreamEnable"
eiger_monitor = "BL24I-EA-EIGER-01:CAM:MonitorEnable"
eiger_datasource = "BL24I-EA-EIGER-01:CAM:DataSource"
eiger_statuspoll = "BL24I-EA-EIGER-01:CAM:ReadStatus.SCAN"
eiger_roi_mode = "BL24I-EA-EIGER-01:CAM:ROIMode"
eiger_ff = "BL24I-EA-EIGER-01:CAM:FlatfieldApplied"
eiger_compress = "BL24I-EA-EIGER-01:CAM:FWCompression"
eiger_compresstype = "BL24I-EA-EIGER-01:CAM:CompressionAlgo"
eiger_od_compress = "BL24I-EA-EIGER-01:OD:Compression"
eiger_od_datatype = "BL24I-EA-EIGER-01:OD:DataType"
eiger_bitdepthrbv = "BL24I-EA-EIGER-01:CAM:BitDepthImage_RBV"
eiger_countmode = "BL24I-EA-EIGER-01:CAM:CountingMode"
eiger_autosum = "BL24I-EA-EIGER-01:CAM:AutoSummation"
eiger_hdrdetail = "BL24I-EA-EIGER-01:CAM:StreamHdrDetail"
eiger_hdrappen = "BL24I-EA-EIGER-01:CAM:StreamHdrAppendix"
eiger_od_capture = "BL24I-EA-EIGER-01:OD:Capture"
eiger_acquire = "BL24I-EA-EIGER-01:CAM:Acquire"
eiger_wavelength = "BL24I-EA-EIGER-01:CAM:Wavelength"
eiger_detdist = "BL24I-EA-EIGER-01:CAM:DetDist"
eiger_beamx = "BL24I-EA-EIGER-01:CAM:BeamX"
eiger_beamy = "BL24I-EA-EIGER-01:CAM:BeamY"
eiger_omegaincr = "BL24I-EA-EIGER-01:CAM:OmegaIncr"
eiger_od_filename_rbv = "BL24I-EA-EIGER-01:OD:FP:FileName_RBV"

# BL24I beamline-specific labelled PVs
me14e_filter = "BL24I-EA-CHIP-01:FILTER"  # to check
me14e_filepath = "BL24I-MO-CHIP-01:filePath"
me14e_chip_name = "BL24I-MO-CHIP-01:chipName"
me14e_chipcapacity = "BL24I-MO-CHIP-01:chipCapacity"
me14e_blockcapacity = "BL24I-MO-CHIP-01:blockCapacity"
me14e_exptime = "BL24I-MO-CHIP-01:expTime"
me14e_dcdetdist = "BL24I-MO-CHIP-01:detDistance"

# General Purpose PV. Now use STEP13
ioc13_gp1 = "BL24I-MO-IOC-13:GP1"
ioc13_gp2 = "BL24I-MO-IOC-13:GP2"
ioc13_gp3 = "BL24I-MO-IOC-13:GP3"
ioc13_gp4 = "BL24I-MO-IOC-13:GP4"
ioc13_gp5 = "BL24I-MO-IOC-13:GP5"
ioc13_gp6 = "BL24I-MO-IOC-13:GP6"
ioc13_gp7 = "BL24I-MO-IOC-13:GP7"
ioc13_gp8 = "BL24I-MO-IOC-13:GP8"
ioc13_gp9 = "BL24I-MO-IOC-13:GP9"
ioc13_gp10 = "BL24I-MO-IOC-13:GP10"
ioc13_gp11 = "BL24I-MO-IOC-13:GP11"
ioc13_gp12 = "BL24I-MO-IOC-13:GP12"
ioc13_gp13 = "BL24I-MO-IOC-13:GP13"
ioc13_gp14 = "BL24I-MO-IOC-13:GP14"
ioc13_gp15 = "BL24I-MO-IOC-13:GP15"
ioc13_gp16 = "BL24I-MO-IOC-13:GP16"
ioc13_gp17 = "BL24I-MO-IOC-13:GP17"
ioc13_gp18 = "BL24I-MO-IOC-13:GP18"
ioc13_gp19 = "BL24I-MO-IOC-13:GP19"
ioc13_gp20 = "BL24I-MO-IOC-13:GP20"
ioc13_gp21 = "BL24I-MO-IOC-13:GP21"
ioc13_gp22 = "BL24I-MO-IOC-13:GP22"
ioc13_gp23 = "BL24I-MO-IOC-13:GP23"
ioc13_gp24 = "BL24I-MO-IOC-13:GP24"
ioc13_gp25 = "BL24I-MO-IOC-13:GP25"
ioc13_gp26 = "BL24I-MO-IOC-13:GP26"
ioc13_gp27 = "BL24I-MO-IOC-13:GP27"
ioc13_gp28 = "BL24I-MO-IOC-13:GP28"
ioc13_gp29 = "BL24I-MO-IOC-13:GP29"
ioc13_gp30 = "BL24I-MO-IOC-13:GP30"
ioc13_gp31 = "BL24I-MO-IOC-13:GP31"
ioc13_gp32 = "BL24I-MO-IOC-13:GP32"
ioc13_gp33 = "BL24I-MO-IOC-13:GP33"
ioc13_gp34 = "BL24I-MO-IOC-13:GP34"
ioc13_gp35 = "BL24I-MO-IOC-13:GP35"
ioc13_gp36 = "BL24I-MO-IOC-13:GP36"
ioc13_gp37 = "BL24I-MO-IOC-13:GP37"
ioc13_gp38 = "BL24I-MO-IOC-13:GP38"
ioc13_gp39 = "BL24I-MO-IOC-13:GP39"
ioc13_gp40 = "BL24I-MO-IOC-13:GP40"
ioc13_gp41 = "BL24I-MO-IOC-13:GP41"
ioc13_gp42 = "BL24I-MO-IOC-13:GP42"
ioc13_gp43 = "BL24I-MO-IOC-13:GP43"
ioc13_gp44 = "BL24I-MO-IOC-13:GP44"
ioc13_gp45 = "BL24I-MO-IOC-13:GP45"
ioc13_gp46 = "BL24I-MO-IOC-13:GP46"
ioc13_gp47 = "BL24I-MO-IOC-13:GP47"
ioc13_gp48 = "BL24I-MO-IOC-13:GP48"
ioc13_gp49 = "BL24I-MO-IOC-13:GP49"
ioc13_gp50 = "BL24I-MO-IOC-13:GP50"
ioc13_gp51 = "BL24I-MO-IOC-13:GP51"
ioc13_gp52 = "BL24I-MO-IOC-13:GP52"
ioc13_gp53 = "BL24I-MO-IOC-13:GP53"
ioc13_gp54 = "BL24I-MO-IOC-13:GP54"
ioc13_gp55 = "BL24I-MO-IOC-13:GP55"
ioc13_gp56 = "BL24I-MO-IOC-13:GP56"
ioc13_gp57 = "BL24I-MO-IOC-13:GP57"
ioc13_gp58 = "BL24I-MO-IOC-13:GP58"
ioc13_gp59 = "BL24I-MO-IOC-13:GP59"
ioc13_gp60 = "BL24I-MO-IOC-13:GP60"
ioc13_gp61 = "BL24I-MO-IOC-13:GP61"
ioc13_gp62 = "BL24I-MO-IOC-13:GP62"
ioc13_gp63 = "BL24I-MO-IOC-13:GP63"
ioc13_gp64 = "BL24I-MO-IOC-13:GP64"
ioc13_gp65 = "BL24I-MO-IOC-13:GP65"
ioc13_gp66 = "BL24I-MO-IOC-13:GP66"
ioc13_gp67 = "BL24I-MO-IOC-13:GP67"
ioc13_gp68 = "BL24I-MO-IOC-13:GP68"
ioc13_gp69 = "BL24I-MO-IOC-13:GP69"
ioc13_gp70 = "BL24I-MO-IOC-13:GP70"
ioc13_gp71 = "BL24I-MO-IOC-13:GP71"
ioc13_gp72 = "BL24I-MO-IOC-13:GP72"
ioc13_gp73 = "BL24I-MO-IOC-13:GP73"
ioc13_gp74 = "BL24I-MO-IOC-13:GP74"
ioc13_gp75 = "BL24I-MO-IOC-13:GP75"
ioc13_gp76 = "BL24I-MO-IOC-13:GP76"
ioc13_gp77 = "BL24I-MO-IOC-13:GP77"
ioc13_gp78 = "BL24I-MO-IOC-13:GP78"
ioc13_gp79 = "BL24I-MO-IOC-13:GP79"
ioc13_gp80 = "BL24I-MO-IOC-13:GP80"
ioc13_gp81 = "BL24I-MO-IOC-13:GP81"
ioc13_gp82 = "BL24I-MO-IOC-13:GP82"
ioc13_gp83 = "BL24I-MO-IOC-13:GP83"
ioc13_gp84 = "BL24I-MO-IOC-13:GP84"
ioc13_gp85 = "BL24I-MO-IOC-13:GP85"
ioc13_gp86 = "BL24I-MO-IOC-13:GP86"
ioc13_gp87 = "BL24I-MO-IOC-13:GP87"
ioc13_gp88 = "BL24I-MO-IOC-13:GP88"
ioc13_gp89 = "BL24I-MO-IOC-13:GP89"
ioc13_gp90 = "BL24I-MO-IOC-13:GP90"
ioc13_gp91 = "BL24I-MO-IOC-13:GP91"
ioc13_gp92 = "BL24I-MO-IOC-13:GP92"
ioc13_gp93 = "BL24I-MO-IOC-13:GP93"
ioc13_gp94 = "BL24I-MO-IOC-13:GP94"
ioc13_gp95 = "BL24I-MO-IOC-13:GP95"
ioc13_gp96 = "BL24I-MO-IOC-13:GP96"
ioc13_gp97 = "BL24I-MO-IOC-13:GP97"
ioc13_gp98 = "BL24I-MO-IOC-13:GP98"
ioc13_gp99 = "BL24I-MO-IOC-13:GP99"
ioc13_gp100 = "BL24I-MO-IOC-13:GP100"
ioc13_gp101 = "BL24I-MO-IOC-13:GP101"
ioc13_gp102 = "BL24I-MO-IOC-13:GP102"
ioc13_gp103 = "BL24I-MO-IOC-13:GP103"
ioc13_gp104 = "BL24I-MO-IOC-13:GP104"
ioc13_gp105 = "BL24I-MO-IOC-13:GP105"
ioc13_gp106 = "BL24I-MO-IOC-13:GP106"
ioc13_gp107 = "BL24I-MO-IOC-13:GP107"
ioc13_gp108 = "BL24I-MO-IOC-13:GP108"
ioc13_gp109 = "BL24I-MO-IOC-13:GP109"
ioc13_gp110 = "BL24I-MO-IOC-13:GP110"
ioc13_gp111 = "BL24I-MO-IOC-13:GP111"
ioc13_gp112 = "BL24I-MO-IOC-13:GP112"
ioc13_gp113 = "BL24I-MO-IOC-13:GP113"
ioc13_gp114 = "BL24I-MO-IOC-13:GP114"
ioc13_gp115 = "BL24I-MO-IOC-13:GP115"
ioc13_gp116 = "BL24I-MO-IOC-13:GP116"
ioc13_gp117 = "BL24I-MO-IOC-13:GP117"
ioc13_gp118 = "BL24I-MO-IOC-13:GP118"
ioc13_gp119 = "BL24I-MO-IOC-13:GP119"
ioc13_gp120 = "BL24I-MO-IOC-13:GP120"
