"""
This package contains the bluesky plan entry points for use with hyperion-blueapi.
The json schema and documentation therein generated by the blueapi /plans endpoint
from this package constitutes the hyperion-blueapi interface to the hyperion supervisor
process.
"""
